"""Actuator command received from WolkAbout IoT Platform."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from typing import Union


@dataclass
class ActuatorCommand:
    """
    Actuator command for reference with value.

    :ivar reference: What actuator is the command for
    :vartype reference: str
    :ivar value: Value to be set
    :vartype value: Union[bool, int, float, str]
    """

    reference: str
    value: Union[bool, int, float, str]
