"""Information about the current status of the firmware update process."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from dataclasses import dataclass
from dataclasses import field
from typing import Optional

from wolk.model.file_management_error_type import FileManagementErrorType
from wolk.model.file_management_status_type import FileManagementStatusType


@dataclass
class FileManagementStatus:
    """
    Contains the status of the file management process.

    :ivar status: The status of the file management process
    :ivartype status: FileManagementStatusType
    :ivar error: The type of error that occurred
    :ivartype error: FileManagementErrorType or None
    """

    status: FileManagementStatusType
    error: Optional[FileManagementErrorType] = field(default=None)
